/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307 USA
 */

#ifndef SNAPSHOT_H
#define SNAPSHOT_H

#include "ns.h"
#include "cfs.h"
#include "cfsrw.h"
#include "cfsr_prot.h"

class snapshot_obj {
public:
  virtual ~snapshot_obj () {}
  virtual bigint root () const = 0;
  virtual bigint addr () const = 0;
  virtual bigint prev () const = 0;
  virtual commit_info branch () const = 0;
  virtual str btag () const = 0;
};

typedef ptr<snapshot_obj> sptr;

class cfsr : public virtual refcount {
private:
  static void lookup 
    (sptr, bigint dir, str name,
     callback<void, bigint, cfsrw_status>::ref cb);
  
  static void getattr
    (sptr, bigint object,
     callback<void, ptr<cfsr_attr>, cfsrw_status>::ref cb);

  static void findver 
    (sptr, bigint object, unsigned rev,
     callback<void, cfsr_findver_res, cfsrw_status>::ref cb);

  static void readdir
    (sptr, bigint dir, uint64 cookie, uint32 count,
     callback<void, ptr<cfs_dirlist>, cfsrw_status>::ref cb);

  static void read
    (sptr, bigint object, uint64 off, uint32 cnt,
     callback<void, ptr<ns_block>, bool, cfsrw_status>::ref cb);

  static void readlink
    (sptr s, bigint file,
     callback<void, str, cfsrw_status>::ref cb);

private:
  friend class cfsrw_impl;
  friend class replica_impl;
  static int init ();

public:
  static void dispatch (ptr<cfsrcall> c);
  static void dispatch (ptr<snapshot_obj> s, ptr<cfsrcall> c) {
    c->_s = s;
    dispatch (c);
  }
};

#endif

