/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef PROJBLOCK_H
#define PROJBLOCK_H

#include "stdinc.h"
#include "pack.h"
#include "ns.h"
#include "vervec.h"

class projblock;

class projblock_constructor {
 private:
   friend class projblock_impl;
   vec<bigint> _head;
   vec<str> _name;
   str _notify;
   void sort ();
   bool parse_old (xin *in);
   bool parse (xin *in);
   void add_entry_no_check (bigint u, str n);
   str serialize ();

public:
  projblock_constructor () { _notify = ""; }
  projblock_constructor (const projblock &);

  bool add_entry (bigint u, str n);
  void set_notify (str x);
  bigint store (ptr<sfspriv> sk, callback<void, ns_stat, bigint>::ref cb);
};

struct projblock_state {
  vervec pkv;
  comvec commits;
  vec<lptr> lptrs;
};

class projblock {
public:
  virtual unsigned nentries () const = 0;
  virtual bigint head (unsigned i) const = 0;
  virtual str name (unsigned i) const = 0;
  virtual str notify () const = 0;
  virtual void fetch (callback<void, ptr<projblock_state>, log_status>::ref cb) = 0;
  virtual void fetch (const vervec& pkv, const comvec& v,
                      callback<void, ptr<projblock_state>, log_status>::ref cb) = 0;
  static ptr<projblock> alloc (callback<void, log_status>::ref cb, bigint addr);
};

#endif

