/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "adb.h"
#include "pastwatch.h"
#include "rxx.h"

int
adb::opendb (str fn, adb_entry_factory *ef, bool create)
{
  dbname = fn;

  str buf = file2str (dbname);
  if (!buf) {
    int fd = open (fn, O_CREAT|O_RDWR, 0600);
    if (fd < 0)
      return -1;
    close (fd);
    return 0;
  }

  vec<str> n;
  split (&n, "\n", buf);
  for (unsigned i=0; i<n.size (); i++) {
    ptr<adb_entry> e = ef->parse_entry (n [i]);
    if (e)
      entries.push_back (e);
  }
  return 0;
}

int
adb::closedb ()
{
  if (modified) {
    strbuf b;
    for (unsigned i=0; i<entries.size (); i++) {
      ptr<adb_entry> e = entries[i];
      e->print_entry (b);
      b << "\n";
    }
    return str2file (dbname, b, 0644, false) ? 0 : -1;
  }
  return 0;
}

