#include "stdinc.h"
#include "pastwatch.h"
#include "opts.h"
#include "mklog.h"
#include "misc_utils.h"
#include "sfskeymgr.h"

static void
create_new_log5 (callback<void, log_status>::ptr cb, ns_stat status, bigint n)
{
  if (status) {
    unlink (arg_logkeyfile.cstr());
    cb (LOGERR_ERR);
    return;
  }
  cb (LOG_OK);
}

static void
create_new_log4 (ptr<sfskey> log_sk, ptr<sfspriv> userkey, ptr<ptr_block> userblock,
		 callback<void, log_status>::ptr cb, cfsrw_status status)
{
  if (status) {
    unlink (arg_logkeyfile.cstr());
    cb (LOGERR_ERR);
    return;
  }

  // Add the log to the userblock
  // mv the temp key file to the final keyfile
  // add the pk to the user block
  bigint lptr_addr = nsobj->pkid (log_sk->key);
  if (!userblock->add_entry (lptr_addr, log_sk->keyname)) {
    warn << "Could not add new key to the user block.  Did you specify "
	 << "a duplicate key label?\n";
    unlink (arg_logkeyfile.cstr());
    cb (LOGERR_MAKEULOGFAIL);
    return;
  }
  userblock->store (userkey, wrap (create_new_log5, cb));
}

static void
create_new_log3 (ptr<projinterface> proj, 
		 ptr<sfspriv> userkey, ptr<ptr_block> userblock,
		 callback<void, log_status>::ptr cb, 
		 log_status status)
{
  if (status) {
    cb (status);
    return;
  }

  warn << "Member block currently has " << userblock->nentries() << " entries:\n";
  userblock->print("  ");

  // Create a keypair for this new log
  str r1;
  u_int nbits = 0;

  ptr<sfskey> log_sk = New refcounted<sfskey> ();
  if (r1 = sfskeygen (log_sk, nbits, NULL, true, KM_NOPWD, KM_NOKBD, false, SFS_RABIN)) {
    warn << "Error generating log key: " << r1 << "\n";
    cb (LOGERR_MAKELKEYFAIL);
  }

  // XXX Check to make sure there are no duplicate labels in the projinterface

  bigint logpk = nsobj->pkid (log_sk->key);
  arg_logkeyfile = 
    pget_replicadir () << "/" << armor64A (arg_projid) << "/logkey." << armor64A(logpk);
  if (r1 = sfskeysave (arg_logkeyfile, log_sk)) {
    warn << "Error saving key to file, " << r1 << "\n";
    cb (LOGERR_MAKELKEYFAIL);
  }

  // Create the new log.  Make sure to add an empty commit
  // record if this is the first log.
  bool first_log = false;
  if (proj->nlogs() < 1) 
    first_log = true;

  ptr<mklog> log_maker = New refcounted<mklog> ();
  log_maker->create_log (log_sk->key, arg_projid, first_log, NULL, false,
			 wrap (create_new_log4, log_sk, userkey, userblock, cb));
}

static void
create_new_log2 (ptr<projinterface> proj, 
		 callback<void, log_status>::ptr cb, 
		 ptr<head_state> st, log_status status) 
{
  if (status) {
    cb (status);
    return;
  }

  ptr<sfspriv> userkey;
  if (!(userkey= read_skey (arg_userkeyfile))) {
    cb (LOGERR_MAKEULOGFAIL);
    return;
  }

  // load the user block so we can add a new entry to the user block
  ptr<ptr_block> userblock = New refcounted<ptr_block> ();
  userblock->read_block (nsobj->pkid (userkey), 
			 wrap (create_new_log3, proj, userkey, userblock, cb));
}

void
create_new_log (callback<void, log_status>::ptr cb) 
{
  // Retrieve project user info so we know whether to initailize the 
  // empty CFS directory.
  ptr<projinterface> proj = New refcounted<projinterface> (arg_projid);
  proj->fetch_heads (wrap (&create_new_log2, proj, cb));
}
