#!/usr/bin/perl -w

my $mklog = "pastmklog";
my $projctl = "pastprojctl";
my $past = "pastwatch";
my $keygen = "pastkeygen";

# All these should be defined after running parse_args 
# (except projectid which can be set after the id is created)
my $projectid;
my $projectlabel;
my $timestamp;
my $pubkey;
my $username;
my $userid;
my $notifyrule;
my $pwroot;
my $projectdir;
my $updfile;
my $cmd;

select(STDOUT); $| = 1;
select(STDERR); $| = 1;

sub usage {
  print "Usage: pastctl cmd args\n";

  print "Member Commands:\n";
  print "  list project -- list members in a project\n";
  print "  join [-P projectID] name project -- create a key pair & member block for a new member\n";
  print "  remove project -- removes information about a project\n";

  print "Project Administrator Commands:\n";
  print "  create [-p projectID] project -- create an empty project\n";
#  print "     -p: use given projectID rather than creating a new project\n";
  print "  addmember [-S timestamp] [-k pubkey] name project -- add a member to an existing project\n";
#  print "     -k: use given pubkey for the member\n";
  print "  notify rule project -- add notification rule to project\n";
  print "\n";
  print "  args: <project> is the label of a project.\n";
  print "        <projectID> is the hash of the project identifier.\n";
  exit;
};

sub parse_label {
    my @args = @_;

    while (my $a = shift(@args)) {
	if (!$projectlabel || $projectlabel eq "") {
	    $projectlabel = $a;
	} else {
	    &usage();
	}
    }
    if (!$projectlabel || $projectlabel eq "") { &usage(); }
}
sub parse_create {
    my @args = @_;

    while (my $a = shift(@args)) {
	if ($a eq "-p") {
	    $projectid = shift(@args);
	} elsif (!$projectlabel || $projectlabel eq "") {
	    $projectlabel = $a;
	    #print "Project label is: $projectlabel\n";
	} else {
	    &usage();
	}
    }
    if (!$projectlabel || $projectlabel eq "") { &usage(); }
}
sub parse_addmember {
    my @args = @_;

    while (my $a = shift(@args)) {
	if ($a eq "-S") {
	    $timestamp = shift(@args);
	} elsif ($a eq "-k") {
	    $pubkey = shift(@args);
	} elsif (!$username || $username eq "") {
	    $username = $a;
	} elsif (!$projectlabel || $projectlabel eq "") {
	    $projectlabel = $a;
	} else {
	    &usage();
	}
    }
    if (!$username || $username eq "") { &usage(); }
    if (!$projectlabel || $projectlabel eq "") { &usage(); }
    #if (!$pubkey || $pubkey eq "") { &usage(); }
}
sub parse_join {
    my @args = @_;

    while (my $a = shift(@args)) {
	if ($a eq "-S") {
	    $timestamp = shift(@args);
	} elsif ($a eq "-P") {
	    $projectid = shift(@args);
	} elsif (!$username || $username eq "") {
	    $username = $a;
	} elsif (!$projectlabel || $projectlabel eq "") {
	    $projectlabel = $a;
	} else {
	    &usage();
	}
    }
    if (!$projectlabel || $projectlabel eq "") { &usage(); }
}
sub parse_notify {
    my @args = @_;

    while (my $a = shift(@args)) {
	if (!$notifyrule || $notifyrule eq "") {
	    $notifyrule = $a;
	} elsif (!$projectlabel || $projectlabel eq "") {
	    $projectlabel = $a;
	} else {
	    &usage();
	}
    }
    if (!$projectlabel || $projectlabel eq "") { &usage(); }
    if (!$notifyrule || $notifyrule eq "") { &usage(); }
}
sub parse_args {
    my @args = @_;

    if (scalar(@args) < 1) { &usage (); }
    $cmd = shift @args;
    
    if ($cmd eq "create") {
	&parse_create(@args);
    } elsif ($cmd eq "addmember") {
	&parse_addmember(@args);
    } elsif ($cmd eq "join") {
	&parse_join(@args);
    } elsif ($cmd eq "remove" || $cmd eq "list") {
	&parse_label(@args);
    } elsif ($cmd eq "notify") {
	&parse_notify(@args);
    } else { 
	&usage (); 
    }

    $pwroot = $ENV{PASTWATCH_ROOT};
    if (!$pwroot || $pwroot eq "") { 
	$pwroot = "$ENV{HOME}/.pastwatch";
    }
    $projectdir = "$pwroot/$projectlabel";
    $updfile = "$pwroot/upd.repositories";
}



# ----------------------------------------------------------
sub getprojid {
  my ($dir) = @_;
  if (-l "$dir/project.id") {
    my $s = readlink "$projectdir/project.id";
    if ($s =~ /^project:([\s\S]+)$/) { return $1; }
    else { die "$projectdir/project.id -- corrupted\n"; }
  }
  else {
    die "cannot find $dir/project.id\n";
  }
}

sub create_member {
    # 1) Create a user keypair
    # 2) Create a user block
    # 3) Print out PK of new member

    my ($projectid, $projectlabel, $username) = @_;

    #&create_proj_dir($pwroot, $projectlabel, $projectid);
    
    my $r = system ("$keygen -KP -l $username /tmp/member.key.$$");
    if ($r) {
	unlink "/tmp/member.key.$$";
	die "cannot create new member's key file\n";
    }

    $r = system ("$projctl -u /tmp/member.key.$$ 1>/tmp/projectctl.$$");
    if ($r) {
	unlink "/tmp/member.key.$$";
	unlink "/tmp/projectctl.$$";
	die "cannot create member block";
    }
    my $userid = `cat /tmp/projectctl.$$`;
    unlink "/tmp/projectctl.$$";
    chomp $userid;

    $r = system ("mv /tmp/member.key.$$ $projectdir/member.$userid");
    if ($r) {
	unlink "/tmp/member.key.$$";
	unlink "/tmp/projectctl.$$";
	die "cannot move member key to $projectdir/member.$userid";
    }
    return $userid;
}

sub create_proj_dir {
    my ($home, $projname, $projid) = @_;

    #print "mkdir $home/$projname\n";

    mkdir ($home,0755);
    if (!mkdir ("$home/$projname",0755) && $! ne "File exists") { 
	die "cannot create $home/$projname: $!\n"; 
    }
    
    my @a = lstat("$home/$projname/project.id");
    if ($#a > 0) {
	die "Symlink $home/$projname/project.id already exists.\n\
You must remove this link before reusing the label: $projname\n";
    }

    if (!symlink "project:$projid", "$home/$projname/project.id") {
	die "Cannot create symlink $home/$projname/project.id: $!\n";
    }
}

sub run_remove {
    my $projectid = &getprojid ($projectdir);
    open (F0, $updfile) || die "can't remove project from $updfile\n";
    open (F1, ">$updfile.$$") || die "can't open $updfile.$$\n";
    while (<F0>) {
	chomp;
	if ($_ ne $projectid) { print F1 "$_\n"; }
    }
    close F1;
    close F0;
    rename ("$updfile.$$", "$updfile");
    system ("rm -rf $projectdir");
}
sub run_create {
    my $r;
    if (-x $projectdir) { 
	die "The directory: $projectdir already exists.\
You must remove it before creating a project with the label: $projectlabel\n"; 
    }

    $tempprojdir = "/tmp/pastprojdir.$$";
    chomp $tempprojdir;
    
    mkdir ($pwroot,0755);
    if (!mkdir ($tempprojdir,0755)) { 
	die "Error while creating temporary directory $tempprojdir: $!\n"; 
    }
    
    if (!$projectid || $projectid eq "") {
	$r = system ("$keygen -KP -l $projectlabel /tmp/project.key.$$");
	if ($r) {
	    system("rm -r $tempprojdir");
	    unlink "/tmp/project.key.$$";
	    die "cannot create project's key file\n";
	}

	$r = system ("$projctl -s /tmp/project.key.$$ 1>/tmp/projectctl.$$");
	if ($r) {
	    system("rm -r $tempprojdir");
	    unlink "/tmp/project.key.$$";
	    unlink "/tmp/projectctl.$$";
	    die "cannot create project\n";
	}

	$projectid = `cat /tmp/projectctl.$$`;
	unlink "/tmp/projectctl.$$";
	chomp $projectid;

	$r = system ("mv /tmp/project.key.$$ $tempprojdir/project.key");
	if ($r) {
	    system("rm -r $tempprojdir");
	    unlink "/tmp/project.key.$$";
	    die "cannot move project's key file into $tempprojdir\n";
	}
    }

    my @a = lstat("$tempprojdir/project.id");
    if ($#a > 0) {
	system("rm -r $tempprojdir");
	die "Symlink $tempprojdir/project.id already exists.\
You must remove this link before creating a new project named $projectlabel\n";
    }
    if (!symlink "project:$projectid", "$tempprojdir/project.id") {
	system("rm -r $tempprojdir");
	die "cannot create symlink $tempprojdir/project.id: $!\n"; 
    }

    print "new project ID: $projectid\n";

    if (!open (F0, ">>$updfile")) {
	system("rm -r $tempprojdir");
	die "can't update $updfile - disconnected commits won't propagate\n";
    }
    print F0 "$projectid\n";
    close F0;
    
    $r = system("mv $tempprojdir $projectdir");
    if ($r) {
	system("rm -r $tempprojdir");
	die "cannot move new project dir into $projectdir\n";
    }
}
sub run_addmember {
    # 1) Add the member PK to the memberlist
    if (!(-e "$projectdir/project.key")) { die "cannot find admin key at $projectdir/project.key\n"; }
    my $projectid = &getprojid ($projectdir);

    if (!$pubkey) {
	$pubkey = &create_member ($projectid, $projectlabel, $username);
	if (!$pubkey) { die "cannot create new member\n"; }
    }

    my $r = system
	("$projctl -s $projectdir/project.key -n $username -a $pubkey $projectid 1> /dev/null");
    if ($r) { die "cannot add member to project\n"; }
    print "new member: $username, $pubkey\n";
}
sub run_join {
    # If the member specified a key, then create a new directory
    if ($projectid && $projectid ne "") {
	&create_proj_dir($pwroot, $projectlabel, $projectid);
    }
    
    my $projectid = &getprojid ($projectdir);
    my $newkey = &create_member ($projectid, $projectlabel, $username);
    print "give $newkey and a member name to project administrator\n";
}
sub run_notify {
    if (!(-e "$projectdir/project.key")) { die "cannot find $projectdir\n"; }
    my $projectid = &getprojid ($projectdir);

    my $r = system ("$projctl -r \"$notifyrule\" -s $projectdir/project.key $projectid 1>/dev/null");
    if ($r) {
	die "cannot add notification rule to project\n";
    }
}
sub run_list {
    my $projectid = &getprojid ($projectdir);
    my $list = `$projctl -p $projectdir/project.id`;
    my @list = split /\n/, $list;
    my $line;
    
    print "project: $projectlabel ($projectid)\n";
    foreach $line (@list) {
	if ($line =~ /log: ([\s\S]+) \(([\s\S]+)\)/) {
	    if (-r "$projectdir/key.$2") { print "  * member: $1 ($2)\n"; }
	    else { print "  log: $1 ($2)\n"; }
	}
	else { print "  $line\n"; }
    }
}
sub main {
    &parse_args(@ARGV);
    
    if ($cmd eq "remove") {
	&run_remove();
    } elsif ($cmd eq "create") {
	&run_create();
    } elsif ($cmd eq "addmember") {
	&run_addmember();
    } elsif ($cmd eq "join") {
	&run_join();
    } elsif ($cmd eq "notify") {
	&run_notify();
    } elsif ($cmd eq "list") {
	&run_list();
    }
}

&main();

